from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.views.generic import TemplateView

from . import backends


class LandingView(TemplateView):
    template_name = "armstrong/donations/landing.html"


class ThanksView(TemplateView):
    template_name = "armstrong/donations/thanks.html"


class DonationFormView(TemplateView):
    template_name = "armstrong/donations/donation.html"
    confirm_template_name = "armstrong/donations/confirm.html"
    form_validation_failed = False
    confirm = False

    @property
    def use_confirm_template(self):
        return not self.form_validation_failed \
                and self.requires_confirmation and self.is_write_request

    def get_template_names(self):
        if self.use_confirm_template:
            return [self.confirm_template_name, ]
        return super(DonationFormView, self).get_template_names()

    @property
    def requires_confirmation(self):
        if not self.confirm:
            return False
        return False if "confirmed" in self.request.POST else True

    @property
    def is_write_request(self):
        return self.request.method in ("POST", "PUT")

    _form_action_url = ""

    @property
    def form_action_url(self):
        return self._form_action_url

    @form_action_url.setter
    def set_form_action_url(self, name):
        self._form_action_url = reverse(name)

    _success_url = None

    @property
    def success_url(self):
        if not self._success_url:
            self._success_url = reverse("donations_thanks")
        return self._success_url

    @success_url.setter
    def set_success_url(self, name):
        self._success_url = reverse(name)

    def add_data_if_write_request(self, kwargs):
        if self.is_write_request:
            kwargs.update({
                "data": self.request.POST,
                "files": self.request.FILES,
            })
        return kwargs

    def get_donation_form_kwargs(self):
        kwargs = {"initial": getattr(self, "donation_form_initial", None)}
        return self.add_data_if_write_request(kwargs)

    def get_donation_form_class(self):
        return backends.get_backend().get_form_class()

    def get_donation_form(self):
        donation_form_class = self.get_donation_form_class()
        return donation_form_class(**self.get_donation_form_kwargs())

    def get_context_data(self, **kwargs):
        donation_form = self.get_donation_form()
        context = {
            "form_action_url": self.form_action_url,
            "donor_form": donation_form.donor_form,
            "donation_form": donation_form,
        }
        context.update(kwargs)
        return context

    def get(self, request, *args, **kwargs):
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        donation_form = self.get_donation_form()
        if not donation_form.is_valid():
            return self.form_is_invalid()
        return self.form_is_valid(donation_form=donation_form)

    def form_is_invalid(self):
        self.form_validation_failed = True
        return self.render_to_response(self.get_context_data())

    def form_is_valid(self, donation_form):
        if self.requires_confirmation:
            context = {"confirmation_required": True}
            context.update(self.get_context_data())
            return self.render_to_response(context)
        donation = donation_form.save()
        response = backends.get_backend().purchase(donation, donation_form)
        if not response["status"]:
            return self.purchase_failed(response)
        return HttpResponseRedirect(self.success_url)

    def purchase_failed(self, backend_response):
        context = {
            "error_msg": "Unable to process payment",
            "reason": backend_response["reason"],
            "response": backend_response["response"],
        }
        context.update(self.get_context_data())
        return self.render_to_response(context)



from armstrong.utils.backends import GenericBackend
from authorize import aim
from authorize import arb
import datetime
from django.conf import settings as django_settings

from . import forms
from . import signals


class AuthorizeNetBackend(object):
    def __init__(self, api_class=None, recurring_api_class=None,
            settings=None, testing=None):
        if api_class is None:
            api_class = aim.Api
        self.api_class = api_class
        if settings is None:
            settings = django_settings
        self.settings = settings
        if recurring_api_class is None:
            recurring_api_class = arb.Api
        self.recurring_api_class = recurring_api_class
        if testing is None:
            testing = getattr(self.settings, "ARMSTRONG_DONATIONS_TESTING",
                    False)
        self.testing = testing

    def get_api(self):
        return self.api_class(self.settings.AUTHORIZE["LOGIN"],
                self.settings.AUTHORIZE["KEY"], delimiter=u"|",
                is_test=self.testing)

    def get_recurring_api(self):
        return self.recurring_api_class(self.settings.AUTHORIZE["LOGIN"],
                self.settings.AUTHORIZE["KEY"], is_test=self.testing)

    def get_form_class(self):
        return forms.AuthorizeDonationForm

    def purchase(self, donation, form):
        result = self.onetime_purchase(donation, form)
        if not result["status"]:
            return result
        if donation.is_repeating:
            response = self.recurring_purchase(donation, form)
            result["recurring_response"] = response
        if result["status"]:
            donation.processed = True
            self.send_successful_purchase(donation, form, result)
        return result

    def send_successful_purchase(self, donation, form, result):
        """
        Called by ``purchase`` after a donation has been succesfully
        processed.

        This function is used to trigger the ``successful_purchase``
        signal while providing a flex point for developers to perform
        an action based on the successful donation prior to the signal
        being sent.
        """
        signals.successful_purchase.send(sender=self, donation=donation,
                form=form, result=result)

    def recurring_purchase(self, donation, form):
        today = datetime.date.today()
        start_date = u"%s" % ((today + datetime.timedelta(days=30))
                .strftime("%Y-%m-%d"))
        api = self.get_recurring_api()
        data = form.get_data_for_charge(donation.donor, recurring=True)
        data.update({
            "amount": donation.amount,
            "interval_unit": arb.MONTHS_INTERVAL,
            "interval_length": u"%d" % donation.donation_type.length,
            "bill_first_name": u"%s" % donation.donor.first_name,
            "bill_last_name": u"%s" % donation.donor.last_name,
            "total_occurrences": donation.donation_type.repeat,
            "start_date": start_date,
        })
        if self.testing:
            data["test_request"] = u"TRUE"
        response = api.create_subscription(**data)
        status = response["messages"]["result_code"]["text_"] == u"Ok"
        return {
            "status": status,
        }

    def onetime_purchase(self, donation, form):
        api = self.get_api()
        data = form.get_data_for_charge(donation.donor)
        donor = donation.donor
        data.update({
            "amount": donation.amount,
            "description": u"Donation: $%d" % donation.amount,
            "first_name": unicode(donor.first_name),
            "last_name": unicode(donor.last_name),

            # TODO: extract and be conditional
            "address": donor.address.address,
            "city": donor.address.city,
            "state": donor.address.state,
            "zip": donor.address.zipcode,
        })
        if self.testing:
            data["test_request"] = u"TRUE"
        response = api.transaction(**data)
        status = response["reason_code"] == u"1"
        return {
            "status": status,
            "reason": response["reason_text"],
            "response": response,
        }


raw_backend = GenericBackend("ARMSTRONG_DONATIONS_BACKEND", defaults=[
    "armstrong.apps.donations.backends.AuthorizeNetBackend",
])

get_backend = raw_backend.get_backend



from authorize import aim, arb
import datetime
from django.conf import settings as django_settings
import fudge
from fudge.inspector import arg
import os
import random
import unittest

from ._utils import TestCase

from .. import backends
from .. import forms
from .. import signals


class AuthorizeNetBackendTestCase(TestCase):
    def get_api_stub(self, response=None, reason_text=None, successful=True):
        if response is None:
            response = {
                "reason_code": u"1" if successful else u"2",
                "reason_text": "Some Random Reason %d" % (
                        random.randint(100, 200)),
            }
        if reason_text:
            response["reason_text"] = reason_text
        api = fudge.Fake()
        api.provides("transaction").returns(response)
        get_api = fudge.Fake().is_callable().returns(api)
        return get_api

    def get_onetime_purchase_stub(self, successful=True):
        onetime_purchase = fudge.Fake().is_callable().returns({
                "status": successful,
        })
        return onetime_purchase

    @property
    def test_settings(self):
        fake = fudge.Fake()
        fake.has_attr(AUTHORIZE={
            # Login/password 2k4NuTk6cS
            "LOGIN": u"5A77vX8HxE",
            "KEY": u"6T29u7p67xKeEW33",
        })
        return fake

    def test_testing_property_defaults_to_false(self):
        backend = backends.AuthorizeNetBackend()
        self.assertFalse(backend.testing)

    def test_testing_uses_settings_for_default_value(self):
        settings = self.test_settings
        settings.ARMSTRONG_DONATIONS_TESTING = True
        backend = backends.AuthorizeNetBackend(settings=settings)
        self.assertTrue(backend.testing)

    def test_testing_kwarg_other_than_none_are_used_in_favor_of_settings(self):
        settings = self.test_settings
        settings.ARMSTRONG_DONATIONS_TESTING = True
        backend = backends.AuthorizeNetBackend(testing=False,
                settings=settings)
        self.assertFalse(backend.testing)

    def test_adds_is_test_to_recurring_api_class_api_if_in_testing_mode(self):
        recurring_api_class = fudge.Fake().expects_call().with_args(arg.any(),
                arg.any(), is_test=True)

        backend = backends.AuthorizeNetBackend(testing=True)
        with fudge.patched_context(backend, "recurring_api_class",
                recurring_api_class):
            backend.get_recurring_api()
        fudge.verify()

    def test_is_test_is_false_for_recurring_api_class_api_by_default(self):
        recurring_api_class = fudge.Fake().expects_call().with_args(arg.any(),
                arg.any(), is_test=False)

        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "recurring_api_class",
                recurring_api_class):
            backend.get_recurring_api()
        fudge.verify()

    def test_adds_is_test_to_api_class_api_if_in_testing_mode(self):
        api_class = fudge.Fake().expects_call().with_args(arg.any(), arg.any(),
                delimiter=arg.any(), is_test=True)

        backend = backends.AuthorizeNetBackend(testing=True)
        with fudge.patched_context(backend, "api_class", api_class):
            backend.get_api()
        fudge.verify()

    def test_is_test_is_false_for_api_class_api_by_default(self):
        api_class = fudge.Fake().expects_call().with_args(arg.any(), arg.any(),
                delimiter=arg.any(), is_test=False)

        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "api_class", api_class):
            backend.get_api()
        fudge.verify()

    def test_settings_defaults_to_django_settings(self):
        backend = backends.AuthorizeNetBackend()
        self.assertEqual(backend.settings, django_settings)

    def test_settings_can_be_injected(self):
        r = random.randint(10000, 20000)
        backend = backends.AuthorizeNetBackend(settings=r)
        self.assertEqual(backend.settings, r)

    def test_api_class_defaults_to_authorize_api(self):
        backend = backends.AuthorizeNetBackend()
        self.assertEqual(backend.api_class, aim.Api)

    def test_api_class_can_be_injected(self):
        r = random.randint(10000, 20000)
        backend = backends.AuthorizeNetBackend(api_class=r)
        self.assertEqual(backend.api_class, r)

    def test_api_instantiates_api_class_with_configured_settings(self):
        random_login = "some random login %d" % random.randint(100, 200)
        random_key = "some random key %d" % random.randint(100, 200)
        random_return = "some random return %d" % random.randint(100, 200)
        settings = fudge.Fake()
        settings.has_attr(AUTHORIZE={
            "LOGIN": random_login,
            "KEY": random_key,
        })
        api_class = fudge.Fake()

        # Note that delimiter is included here because authorize's code
        # can't even keep track of what deliminter it wants to use!
        (api_class.expects_call()
                .with_args(random_login, random_key, delimiter=arg.any(),
                        is_test=arg.any())
                .returns(random_return))
        fudge.clear_calls()

        backend = backends.AuthorizeNetBackend(api_class=api_class,
                settings=settings)
        result = backend.get_api()
        self.assertEqual(result, random_return)
        fudge.verify()

    def test_recurring_api_class_defaults_to_authorize_api(self):
        backend = backends.AuthorizeNetBackend()
        self.assertEqual(backend.recurring_api_class, arb.Api)

    def test_recurring_api_class_can_be_injected(self):
        r = random.randint(10000, 20000)
        backend = backends.AuthorizeNetBackend(recurring_api_class=r)
        self.assertEqual(backend.recurring_api_class, r)

    def test_get_recurring_api_instantiates_with_configured_settings(self):
        random_login = "some random login %d" % random.randint(100, 200)
        random_key = "some random key %d" % random.randint(100, 200)
        random_return = "some random return %d" % random.randint(100, 200)
        settings = fudge.Fake().has_attr(
            AUTHORIZE={
                "LOGIN": random_login,
                "KEY": random_key,
        })

        recurring_api_class = (fudge.Fake().expects_call()
                .with_args(random_login, random_key, is_test=arg.any())
                .returns(random_return))
        fudge.clear_calls()

        backend = backends.AuthorizeNetBackend(settings=settings,
                    recurring_api_class=recurring_api_class)
        result = backend.get_recurring_api()
        self.assertEqual(result, random_return)
        fudge.verify()

    def test_get_form_returns_credit_card_form(self):
        backend = backends.get_backend()
        self.assertEqual(backend.get_form_class(),
                forms.AuthorizeDonationForm)

    def test_dispatches_to_authorize_to_create_transaction(self):
        donation, donation_form = self.random_donation_and_form

        api = fudge.Fake("api")
        api.expects("transaction").with_args(
                amount=donation.amount,
                card_num=donation_form.cleaned_data["card_number"],
                card_code=donation_form.cleaned_data["ccv_code"],
                exp_date=u"%02d-%04d" % (
                        int(donation_form.cleaned_data["expiration_month"]),
                        int(donation_form.cleaned_data["expiration_year"])),
                description=u"Donation: $%d" % donation.amount,
                first_name=unicode(donation.donor.first_name),
                last_name=unicode(donation.donor.last_name),
                address=donation.donor.address.address,
                city=donation.donor.address.city,
                state=donation.donor.address.state,
                zip=donation.donor.address.zipcode,
        ).returns({"reason_code": u"1", "reason_text": u"Some random Reason"})
        get_api = fudge.Fake().expects_call().returns(api)

        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", get_api):
            backend.purchase(donation, donation_form)
        fudge.verify()

    def test_adds_test_request_to_kwargs_if_in_testing_mode(self):
        donation, donation_form = self.random_donation_and_form

        api = fudge.Fake("api")
        api.expects("transaction").with_args(
                amount=donation.amount,
                card_num=donation_form.cleaned_data["card_number"],
                card_code=donation_form.cleaned_data["ccv_code"],
                exp_date=u"%02d-%04d" % (
                        int(donation_form.cleaned_data["expiration_month"]),
                        int(donation_form.cleaned_data["expiration_year"])),
                description=u"Donation: $%d" % donation.amount,
                first_name=unicode(donation.donor.first_name),
                last_name=unicode(donation.donor.last_name),
                address=donation.donor.address.address,
                city=donation.donor.address.city,
                state=donation.donor.address.state,
                zip=donation.donor.address.zipcode,
                test_request=u"TRUE",
        ).returns({"reason_code": u"1", "reason_text": u"Some random Reason"})
        get_api = fudge.Fake().expects_call().returns(api)

        backend = backends.AuthorizeNetBackend(testing=True)
        with fudge.patched_context(backend, "get_api", get_api):
            backend.purchase(donation, donation_form)
        fudge.verify()

    @unittest.skipIf(os.environ.get("FULL_TEST_SUITE", False) != "1",
            "Only run when FULL_TEST_SUITE env is set")
    def test_can_communicate_with_real_authorize_backend(self):
        class TestableApi(aim.Api):
            def __init__(self, *args, **kwargs):
                kwargs["is_test"] = True
                super(TestableApi, self).__init__(*args, **kwargs)

            def transaction(self, **kwargs):
                kwargs["test_request"] = u"TRUE"
                return super(TestableApi, self).transaction(**kwargs)

        donation, donation_form = self.random_donation_and_form
        donation_form.data["card_number"] = u"4222222222222"  # Set to test CC
        donation.amount = 1
        backend = backends.AuthorizeNetBackend(api_class=TestableApi,
                settings=self.test_settings)
        result = backend.purchase(donation, donation_form)
        try:
            self.assertTrue(result["status"], msg="This ")
        except AssertionError:
            # This is a known issue where Authorize.net randomly returns
            # a bad response in test mode.  Yup, you read that correctly.
            # A system designed to process your money can't actually figure
            # out how to run a test server in a reliable wayself.
            self.assertEqual(result["reason"],
                    u"(TESTMODE) The credit card number is invalid.",
                    msg="Authorize.net really has failed us")

    @unittest.skipIf(os.environ.get("FULL_TEST_SUITE", False) != "1",
            "Only run when FULL_TEST_SUITE env is set")
    def test_can_communicate_with_real_authorize_backend_for_recurring(self):
        onetime_purchase = fudge.Fake().is_callable().returns({"status": True})

        class TestableApi(arb.Api):
            def __init__(self, *args, **kwargs):
                kwargs["is_test"] = True
                super(TestableApi, self).__init__(*args, **kwargs)

            def create_subscription(self, **kwargs):
                kwargs["test_request"] = u"TRUE"
                return super(TestableApi, self).create_subscription(**kwargs)

        donation, donation_form = self.random_donation_and_form
        donation_form.data["card_number"] = u"4222222222222"  # Set to test CC
        donation.donation_type = self.random_monthly_type
        donation.amount = 1
        backend = backends.AuthorizeNetBackend(recurring_api_class=TestableApi,
                settings=self.test_settings)
        with fudge.patched_context(backend, "onetime_purchase",
                onetime_purchase):
            result = backend.purchase(donation, donation_form)
        try:
            self.assertTrue(result["status"])
        except AssertionError:
            # This is a known issue where Authorize.net randomly returns
            # a bad response in test mode.  Yup, you read that correctly.
            # A system designed to process your money can't actually figure
            # out how to run a test server in a reliable wayself.
            self.assertEqual(result["reason"],
                    u"(TESTMODE) The credit card number is invalid.",
                    msg="Authorize.net really has failed us")

    def test_mark_donation_as_processed(self):
        donation, donation_form = self.random_donation_and_form
        self.assertFalse(donation.processed, msg="sanity check")
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", self.get_api_stub()):
            backend.purchase(donation, donation_form)

        self.assertTrue(donation.processed)

    def test_donation_processed_is_false_if_not_successfully_charged(self):
        stub = self.get_onetime_purchase_stub(successful=False)
        donation, donation_form = self.random_donation_and_form
        backend = backends.AuthorizeNetBackend()
        with stub_onetime_purchase(backend, stub):
            backend.purchase(donation, donation_form)

        self.assertFalse(donation.processed)

    def test_purchase_returns_true_status_if_successful(self):
        donation, donation_form = self.random_donation_and_form
        get_api = self.get_api_stub()
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", get_api):
            response = backend.purchase(donation, donation_form)
        self.assertTrue(response["status"])

    def test_purchase_returns_false_status_if_successful(self):
        donation, donation_form = self.random_donation_and_form
        get_api = self.get_api_stub(successful=False)
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", get_api):
            response = backend.purchase(donation, donation_form)
        self.assertFalse(response["status"])

    def test_returned_dict_contains_response_reason_text_as_reason(self):
        random_text = "Some Random Text (%d)" % random.randint(1000, 2000)
        donation, donation_form = self.random_donation_and_form
        get_api = self.get_api_stub(reason_text=random_text)
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", get_api):
            result = backend.purchase(donation, donation_form)
        self.assertTrue("reason" in result, msg="sanity check")
        self.assertEqual(result["reason"], random_text)

    def test_returned_dict_contains_raw_response_dict(self):
        random_response = {
            "reason_code": u"%d" % random.randint(100, 200),
            "reason_text": "Some Random Reason %d" % random.randint(100, 200),
        }
        get_api = self.get_api_stub(response=random_response)
        donation, donation_form = self.random_donation_and_form
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", get_api):
            result = backend.purchase(donation, donation_form)
        self.assertTrue("response" in result, msg="sanity check")
        self.assertEqual(result["response"], random_response)

    def test_calls_to_recurring_donation_if_donation_is_recurring(self):
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_monthly_type
        today = datetime.date.today()
        start_date = u"%s" % ((today + datetime.timedelta(days=30))
                .strftime("%Y-%m-%d"))

        recurring_api = fudge.Fake()
        expiration_date = u"%(expiration_year)s-%(expiration_month)s" % (
                donation_form.cleaned_data)
        recurring_api.expects("create_subscription").with_args(
                amount=donation.amount,
                interval_unit=arb.MONTHS_INTERVAL,
                interval_length=u"1",
                card_number=donation_form.cleaned_data["card_number"],
                card_code=donation_form.cleaned_data["ccv_code"],
                expiration_date=expiration_date,
                bill_first_name=u"%s" % donation.donor.first_name,
                bill_last_name=u"%s" % donation.donor.last_name,
                total_occurrences=donation.donation_type.repeat,
                start_date=start_date,
        ).returns({"messages": {"result_code": {"text_": u"Ok"}}})

        fake = fudge.Fake().expects_call().returns(recurring_api)
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", self.get_api_stub()):
            with fudge.patched_context(backend, "get_recurring_api", fake):
                backend.purchase(donation, donation_form)
        fudge.verify()

    def test_changes_interval_length_to_12_on_yearly_donation(self):
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_yearly_type
        today = datetime.date.today()
        start_date = u"%s" % ((today + datetime.timedelta(days=30))
                .strftime("%Y-%m-%d"))

        recurring_api = fudge.Fake()
        expiration_date = u"%(expiration_year)s-%(expiration_month)s" % (
                donation_form.cleaned_data)
        recurring_api.expects("create_subscription").with_args(
                amount=donation.amount,
                interval_unit=arb.MONTHS_INTERVAL,
                interval_length=u"12",
                card_number=donation_form.cleaned_data["card_number"],
                card_code=donation_form.cleaned_data["ccv_code"],
                expiration_date=expiration_date,
                bill_first_name=u"%s" % donation.donor.first_name,
                bill_last_name=u"%s" % donation.donor.last_name,
                total_occurrences=donation.donation_type.repeat,
                start_date=start_date,
        ).returns({"messages": {"result_code": {"text_": u"Ok"}}})

        fake = fudge.Fake().expects_call().returns(recurring_api)
        backend = backends.AuthorizeNetBackend()
        with fudge.patched_context(backend, "get_api", self.get_api_stub()):
            with fudge.patched_context(backend, "get_recurring_api", fake):
                backend.purchase(donation, donation_form)
        fudge.verify()

    def test_adds_test_request_to_recurring_if_in_testing_mode(self):
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_monthly_type
        today = datetime.date.today()
        start_date = u"%s" % ((today + datetime.timedelta(days=30))
                .strftime("%Y-%m-%d"))

        recurring_api = fudge.Fake()
        expiration_date = u"%(expiration_year)s-%(expiration_month)s" % (
                donation_form.cleaned_data)
        recurring_api.expects("create_subscription").with_args(
                amount=donation.amount,
                interval_unit=arb.MONTHS_INTERVAL,
                interval_length=u"1",
                card_number=donation_form.cleaned_data["card_number"],
                card_code=donation_form.cleaned_data["ccv_code"],
                expiration_date=expiration_date,
                bill_first_name=u"%s" % donation.donor.first_name,
                bill_last_name=u"%s" % donation.donor.last_name,
                total_occurrences=donation.donation_type.repeat,
                start_date=start_date,
                test_request=u"TRUE",
        ).returns({"messages": {"result_code": {"text_": u"Ok"}}})

        fake = fudge.Fake().expects_call().returns(recurring_api)
        backend = backends.AuthorizeNetBackend(testing=True)
        with fudge.patched_context(backend, "get_api", self.get_api_stub()):
            with fudge.patched_context(backend, "get_recurring_api", fake):
                backend.purchase(donation, donation_form)
        fudge.verify()

    def test_calls_transaction_prior_to_subscription(self):
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_monthly_type

        recurring_purchase = (fudge.Fake().expects_call()
                    .with_args(donation, donation_form))
        onetime_purchase = (fudge.Fake().expects_call()
                    .with_args(donation, donation_form)
                    .returns({"status": True}))

        backend = backends.AuthorizeNetBackend()
        with stub_recurring_purchase(backend, recurring_purchase):
            with stub_onetime_purchase(backend, onetime_purchase):
                backend.purchase(donation, donation_form)
        fudge.verify()

    def test_does_not_call_recurring_purchase_on_failed_onetime_purchase(self):
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_monthly_type

        recurring_purchase = fudge.Fake()
        onetime_purchase = (fudge.Fake().expects_call()
                    .with_args(donation, donation_form)
                    .returns({"status": False}))

        backend = backends.AuthorizeNetBackend()
        with stub_recurring_purchase(backend, recurring_purchase):
            with stub_onetime_purchase(backend, onetime_purchase):
                backend.purchase(donation, donation_form)
        fudge.verify()

    def test_adds_recurring_response_to_return_on_failure(self):
        random_return = random.randint(1000, 2000)
        donation, donation_form = self.random_donation_and_form
        donation.donation_type = self.random_monthly_type

        recurring_purchase = (fudge.Fake().is_callable()
                .returns(random_return))
        onetime_purchase = (fudge.Fake().is_callable()
                .returns({"status": True}))

        backend = backends.AuthorizeNetBackend()
        with stub_recurring_purchase(backend, recurring_purchase):
            with stub_onetime_purchase(backend, onetime_purchase):
                result = backend.purchase(donation, donation_form)
        self.assertTrue("recurring_response" in result)
        self.assertEqual(result["recurring_response"], random_return)

    def test_fires_donation_complete_on_successful_charge(self):
        donation, form = self.random_donation_and_form
        result = {
            "status": True,
            "random": random.randint(100, 200),
        }
        onetime_purchase = (fudge.Fake().is_callable()
                .returns(result))

        backend = backends.AuthorizeNetBackend()
        signal = (fudge.Fake().expects_call()
                .with_args(sender=backend, donation=donation,
                        signal=arg.any(), form=form, result=result))
        signals.successful_purchase.connect(signal)
        with stub_onetime_purchase(backend, onetime_purchase):
            backend.purchase(donation, form)

        fudge.verify()
        signals.successful_purchase.disconnect(signal)

from contextlib import contextmanager


@contextmanager
def stub_recurring_purchase(backend, recurring_purchase):
    with fudge.patched_context(backend, "recurring_purchase",
            recurring_purchase):
        yield


@contextmanager
def stub_onetime_purchase(backend, onetime_purchase):
    with fudge.patched_context(backend, "onetime_purchase",
            onetime_purchase):
        yield



from django.utils.translation import ugettext as _

default_labels = {
    "donation.label.anonymous": "Keep my donation anonymous",
}


def get(key):
    """

    .. todo:: Make the dictionary configurable
    """
    if key in default_labels:
        return _(default_labels[key])
    return ""



from django import forms
from django.conf import settings
from django.contrib.auth.models import User

from .constants import MONTH_CHOICES
from .constants import YEAR_CHOICES
from .constants import MAILING_SAME_AS_BILLING

from . import models
from . import text

state_kwargs_fields = {}
if hasattr(settings, "ARMSTRONG_INITIAL_STATE"):
    state_kwargs_fields["initial"] = settings.ARMSTRONG_INITIAL_STATE


class BaseDonationForm(forms.Form):
    first_name = forms.CharField()
    last_name = forms.CharField()
    amount = forms.DecimalField(decimal_places=2)
    attribution = forms.CharField(required=False,
            help_text=text.get("donation.help_text.attribution"))
    anonymous = forms.BooleanField(required=False,
            label=text.get("donation.label.anonymous"))

    def __init__(self, data=None, prefix=None, *args, **kwargs):
        # TODO: provide custom prefixes to each sub-form
        self.mailing_same_as_billing = False
        # TODO: make this work with prefixes
        if (data and MAILING_SAME_AS_BILLING in data and
                data[MAILING_SAME_AS_BILLING]):
            self.mailing_same_as_billing = True
        self.donor_form = self.get_donor_form(data=data, prefix=prefix,
                *args, **kwargs)
        billing_address_prefix = "%sbilling" % ("" if not prefix else prefix)
        self.billing_address_form = self.get_billing_address_form(
                data=data, prefix=billing_address_prefix, empty_permitted=True,
                *args, **kwargs)
        self.billing_address_form.empty_permitted = True
        mailing_address_prefix = "%smailing" % ("" if not prefix else prefix)
        self.mailing_address_form = self.get_mailing_address_form(
                data=data, prefix=mailing_address_prefix, *args, **kwargs)
        self.mailing_address_form.empty_permitted = True
        super(BaseDonationForm, self).__init__(data=data, prefix=prefix,
                *args, **kwargs)

    def get_billing_address_form(self, *args, **kwargs):
        return DonorAddressForm(*args, **kwargs)

    def get_mailing_address_form(self, *args, **kwargs):
        return DonorAddressForm(*args, **kwargs)

    def get_donor_form(self, *args, **kwargs):
        return DonorForm(*args, **kwargs)

    def get_donation_kwargs(self):
        if not self.is_valid(donation_only=True):
            return {}
        return {
            "amount": self.cleaned_data["amount"],
            "attribution": self.cleaned_data["attribution"],
            "anonymous": bool(self.cleaned_data["anonymous"]),
        }

    def is_valid(self, donation_only=False):
        donation_is_valid = super(BaseDonationForm, self).is_valid()
        if not donation_is_valid and "amount" in self.errors:
            donation_type_field = self.add_prefix("donation_type_pk")
            if donation_type_field in self.data:
                donation_type_pk = self.data[donation_type_field]
                try:
                    models.DonationTypeOption.objects.get(pk=donation_type_pk)
                    del self.errors["amount"]
                    donation_is_valid = len(self.errors) is 0
                except models.DonationTypeOption.DoesNotExist:
                    donation_is_valid = False

        if donation_only:
            return donation_is_valid
        mailing_address_validity = self.billing_address_form.is_valid() \
                if self.mailing_same_as_billing \
                else self.mailing_address_form.is_valid()
        return all([
            donation_is_valid,
            self.donor_form.is_valid(),
            self.billing_address_form.is_valid(),
            mailing_address_validity,
        ])

    # TODO: support commit=True?
    def save(self, **kwargs):
        donation = models.Donation(**self.get_donation_kwargs())
        promo_code_field_name = self.add_prefix("promo_code")
        if (promo_code_field_name in self.data
                and self.data[promo_code_field_name]):
            donation.code = models.PromoCode.objects.get(
                    code=self.data[promo_code_field_name])
        if self.add_prefix("donation_type_pk") in self.data:
            donation.donation_type = models.DonationTypeOption.objects.get(
                    pk=self.data[self.add_prefix("donation_type_pk")])
        donor = self.donor_form.save(commit=False)
        try:
            user = User.objects.get(pk=self.data[self.add_prefix("user_pk")])
            donor.user = user
            if not donor.email:
                donor.email = user.email
        except (KeyError, ValueError):
            pass
        if self.billing_address_form.is_valid():
            donor.address = self.billing_address_form.save()
            donor.mailing_address = self.mailing_address_form.save() if \
                    not self.mailing_same_as_billing else donor.address
        donor.save()
        donation.donor = donor
        donation.save()
        return donation


class CreditCardDonationForm(BaseDonationForm):
    """
    Form representing the credit card donation process for Authorize.net

    It's possible that this can work with other backends, but it has not
    been tested.

    Each specific Donation form should create what is necessary for their
    backend's implemenation of `purchase`.

    .. todo:: Add widget that is smart for expiration dates
    """
    card_number = forms.CharField()
    ccv_code = forms.CharField()
    expiration_month = forms.ChoiceField(choices=MONTH_CHOICES)
    expiration_year = forms.ChoiceField(choices=YEAR_CHOICES)

    def get_data_for_charge(self, donor, **kwargs):
        raise NotImplementedError


class AuthorizeDonationForm(CreditCardDonationForm):
    def get_data_for_charge(self, donor, recurring=False):
        self.is_valid()
        card_number = "card_num" if not recurring else "card_number"
        data = {
            card_number: self.cleaned_data["card_number"],
            "card_code": self.cleaned_data["ccv_code"],
        }

        if recurring:
            data["expiration_date"] = u"%04d-%02d" % (
                    int(self.cleaned_data["expiration_year"]),
                    int(self.cleaned_data["expiration_month"]))
        else:
            data["exp_date"] = u"%02d-%04d" % (
                    int(self.cleaned_data["expiration_month"]),
                    int(self.cleaned_data["expiration_year"]))
        return data


class DonorForm(forms.ModelForm):
    class Meta:
        model = models.Donor
        excludes = ("address", "mailing_address", )


class DonorAddressForm(forms.ModelForm):
    address = forms.CharField(widget=forms.Textarea)

    class Meta:
        model = models.DonorAddress



"""

.. todo:: Fix issue if this were to stay in memory across a new year
"""
import datetime
now = datetime.datetime.now()

MONTHS = ["%02d" % (i + 1) for i in range(12)]
MONTH_CHOICES = zip(MONTHS, MONTHS)

YEARS = range(now.year, now.year + 10)
YEAR_CHOICES = zip(YEARS, YEARS)

MAILING_SAME_AS_BILLING = "mailing_same_as_billing"
